"""Types and functionality relating to queries"""
import typing as t
from functools import partial

from .clients import send
from .compat import event_loop, func_from_method, urllib_request
from .http import basic_auth

__all__ = [
    'Query',
    'execute',
    'execute_async',
    'executor',
    'async_executor',
    'related',
]

T = t.TypeVar('T')


def _identity(obj):
    return obj


class Query(t.Generic[T]):
    """Abstract base class for query-like objects.
    Any object whose :meth:`~object.__iter__`
    returns a :class:`~snug.http.Request`/:class:`~snug.http.Response`
    generator implements it.

    Note
    ----
    :term:`Generator iterator`\\s themselves also implement this interface
    (i.e. :meth:`~object.__iter__` returns the generator itself).

    Note
    ----
    Query is a :class:`~typing.Generic`.
    This means you may write ``Query[<returntype>]``
    as a descriptive type annotation.

    For example: ``Query[bool]`` indicates a query which returns a boolean.

    Examples
    --------

    Creating a query from a generator function:

    >>> def repo(name: str, owner: str) -> snug.Query[dict]:
    ...    \"\"\"a repo lookup by owner and name\"\"\"
    ...    req = snug.GET(f'https://api.github.com/repos/{owner}/{name}')
    ...    response = yield req
    ...    return json.loads(response.content)
    ...
    >>> query = repo('Hello-World', owner='octocat')

    Creating a class-based query:

    >>> # actually subclassing `Query` is not required
    >>> class repo(snug.Query[dict]):
    ...    \"\"\"a repository lookup by owner and name\"\"\"
    ...    def __init__(self, name: str, owner: str):
    ...        self.name, self.owner = name, owner
    ...
    ...    def __iter__(self):
    ...        owner, name = self.owner, self.name
    ...        request = snug.GET(
    ...            f'https://api.github.com/repos/{owner}/{name}')
    ...        response = yield request
    ...        return json.loads(response.content)
    ...
    >>> query = repo('Hello-World', owner='octocat')

    Creating a fully customized query (for rare cases):

    >>> # actually subclassing `Query` is not required
    >>> class asset_download(snug.Query):
    ...     \"\"\"streaming download of a repository asset.
    ...     Can only be executed with particular clients\"\"\"
    ...     def __init__(self, repo_name, repo_owner, id):
    ...         self.request = snug.GET(
    ...             f'https://api.github.com/repos/{repo_owner}'
    ...             f'/{repo_name}/releases/assets/{id}',
    ...             headers={'Accept': 'application/octet-stream'})
    ...
    ...     def __execute__(self, client, authenticate):
    ...         \"\"\"execute, returning a streaming requests response\"\"\"
    ...         assert isinstance(client, requests.Session)
    ...         req = authenticate(self.request)
    ...         return client.request(req.method, req.url,
    ...                               data=req.content,
    ...                               params=req.params,
    ...                               headers=req.headers,
    ...                               stream=True)
    ...
    ...     async def __execute_async__(self, client, authenticate):
    ...         ...
    ...
    >>> query = asset_download('hub', rep_owner='github', id=4187895)
    >>> response = snug.execute(download, client=requests.Session())
    >>> for chunk in response.iter_content():
    ...    ...
    """
    def __iter__(self):
        """A generator iterator which resolves the query

        Returns
        -------
        ~typing.Generator[Request, Response, T]
        """
        raise NotImplementedError()

    def __execute__(self, client, auth):
        """Default execution logic for a query,
        which uses the query's :meth:`~Query.__iter__`.
        May be overriden for full control of query execution,
        at the cost of reusability.

        .. versionadded:: 1.1

        Note
        ----
        You shouldn't need to override this method, except in rare cases
        where implementing :meth:`Query.__iter__` does not suffice.

        Parameters
        ----------
        client
            the client instance passed to :func:`execute`
        auth: ~typing.Callable[[Request], Request]
            a callable to authenticate a :class:`~snug.http.Request`

        Returns
        -------
        T
            the query result
        """
        gen = iter(self)
        request = next(gen)
        while True:
            response = send(client, auth(request))
            try:
                request = gen.send(response)
            except StopIteration as e:
                return e.args[0]

    # this method is overwritten when importing the _async module (py3 only)
    def __execute_async__(self, client, auth):
        raise NotImplementedError('python 3+ required to execute async')


_default_execute_method = func_from_method(Query.__execute__)


class related(object):
    """Decorate classes to make them callable as methods.
    This can be used to implement related queries
    through nested classes.

    Example
    -------

    >>> class Foo:
    ...     @related
    ...     class Bar:
    ...         def __init__(self, foo, qux):
    ...             self.the_foo, self.qux = foo, qux
    ...
    >>> f = Foo()
    >>> b = p.Bar(qux=5)
    >>> isinstance(b, Foo.Bar)
    True
    >>> b.the_foo is f
    True
    """
    def __init__(self, cls):
        self._cls = cls

    def __get__(self, obj, objtype=None):
        return self._cls if obj is None else partial(self._cls, obj)


def _make_auth(auth):
    if auth is None:
        return _identity
    elif callable(auth):
        return auth
    else:
        return basic_auth(auth)


def execute(query, auth=None, client=urllib_request.build_opener()):
    """Execute a query, returning its result

    Parameters
    ----------
    query: Query[T]
        The query to resolve
    auth: ~typing.Tuple[str, str] \
        or ~typing.Callable[[Request], Request] or None
        This may be:

        * A (username, password)-tuple for basic authentication
        * A callable to authenticate requests.
        * ``None`` (no authentication)
    client
        The HTTP client to use.
        Its type must have been registered
        with :func:`~snug.clients.send`.
        If not given, the built-in :mod:`urllib` module is used.

    Returns
    -------
    T
        the query result
    """
    exec_fn = getattr(type(query), '__execute__', _default_execute_method)
    return exec_fn(query, client, _make_auth(auth))


def execute_async(query, auth=None, client=event_loop):
    """Execute a query asynchronously, returning its result

    Parameters
    ----------
    query: Query[T]
        The query to resolve
    auth: ~typing.Tuple[str, str] \
        or ~typing.Callable[[Request], Request] or None
        This may be:

        * A (username, password)-tuple for basic authentication
        * A callable to authenticate requests.
        * ``None`` (no authentication)
    client
        The HTTP client to use.
        Its type must have been registered
        with :func:`~snug.clients.send_async`.
        If not given, the built-in :mod:`asyncio` module is used.

    Returns
    -------
    T
        the query result

    Note
    ----
    The default client is very rudimentary.
    Consider using a :class:`aiohttp.ClientSession` instance as ``client``.
    """
    exc_fn = getattr(type(query), '__execute_async__', Query.__execute_async__)
    return exc_fn(query, client, _make_auth(auth))


def executor(**kwargs):
    """Create a version of :func:`execute` with bound arguments.

    Parameters
    ----------
    **kwargs
        arguments to pass to :func:`execute`

    Returns
    -------
    ~typing.Callable[[Query[T]], T]
        an :func:`execute`-like function
    """
    return partial(execute, **kwargs)


def async_executor(**kwargs):
    """Create a version of :func:`execute_async` with bound arguments.

    Parameters
    ----------
    **kwargs
        arguments to pass to :func:`execute_async`

    Returns
    -------
    ~typing.Callable[[Query[T]], ~typing.Awaitable[T]]
        an :func:`execute_async`-like function
    """
    return partial(execute_async, **kwargs)



"""Tools for creating paginated queries

.. versionadded:: 1.2
"""
import abc
import typing as t
from operator import attrgetter

from .compat import PY3
from .query import Query, async_executor, executor

__all__ = [
    'paginated',
    'Page',
    'Pagelike',
]

T = t.TypeVar('T')

if PY3:
    from ._async import AsyncPaginator
    _AsyncIterator = t.AsyncIterator
else:  # pragma: no cover
    _AsyncIterator = t.Iterator


class Pagelike(t.Generic[T]):
    """Abstract base class for page-like objects.
    Any object implementing the attributes
    :py:attr:`~Pagelike.content` and :py:attr:`~Pagelike.next_query`
    implements this interface.
    A query returning such an object may be :class:`paginated`.

    Note
    ----
    Pagelike is a :class:`~typing.Generic`.
    This means you may write ``Pagelike[<type-of-content>]``
    as a descriptive type annotation.

    For example: ``Pagelike[List[str]]`` indicates a page-like
    object whose ``content`` is a list of strings.
    """
    __slots__ = ()

    @abc.abstractproperty
    def content(self):
        """The contents of the page.

        Returns
        -------
        T
            The page content.
        """
        raise NotImplementedError()

    @abc.abstractproperty
    def next_query(self):
        """The query to retrieve the next page,
        or ``None`` if there is no next page.

        Returns
        -------
        ~snug.Query[Pagelike[T]]] or None
            The next query.
        """
        raise NotImplementedError()


class Page(Pagelike[T]):
    """A simple :class:`Pagelike` object

    Parameters
    ----------
    content: T
        The page content.
    next_query: ~snug.Query[Pagelike[T]]] or None
        The query to retrieve the next page.
    """
    __slots__ = '_content', '_next_query'

    def __init__(self, content, next_query=None):
        self._content, self._next_query = content, next_query

    content = property(attrgetter('_content'))
    next_query = property(attrgetter('_next_query'))

    def __repr__(self):
        return 'Page({})'.format(self._content)


class paginated(Query[t.Union[t.Iterator[T], _AsyncIterator[T]]]):
    """A paginated version of a query.
    Executing it returns an :term:`iterator`
    or :term:`async iterator <asynchronous iterator>`.

    If the wrapped query is reusable,
    the paginated query is also reusable.

    Parameters
    ----------
    query: Query[Pagelike[T]]
        The query to paginate.
        This query must return a :class:`Pagelike` object.

    Note
    ----
    Async iterators were introduced in
    `PEP 492 <https://www.python.org/dev/peps/pep-0492>`_.
    Therefore, async execution of :class:`paginated`
    queries is only supported on python 3.5.2+.

    Example
    -------

    .. code-block:: python

        def foo_page(...) -> Query[Pagelike[Foo]]  # example query
            ...
            return Page(...)

        query = paginated(foo_page(...))

        for foo in execute(query):
            ...

        async for foo in execute_async(query):  # python 3.5.2+ only
            ...
    """
    __slots__ = '_query'

    def __init__(self, query):
        self._query = query

    def __execute__(self, client, auth):
        """Execute the paginated query.

        Returns
        -------
        ~typing.Iterator[T]
            An iterator yielding page content.
        """
        return Paginator(self._query, executor(client=client, auth=auth))

    if PY3:
        def __execute_async__(self, client, auth):
            """Execute the paginated query asynchronously.

            Note
            ----
            This method does not need to be awaited.

            Returns
            -------
            ~typing.AsyncIterator[T]
                An asynchronous iterator yielding page content.
            """
            return AsyncPaginator(self._query,
                                  async_executor(client=client, auth=auth))

    else:  # pragma: no cover
        def __execute_async__(self, client, auth):
            raise NotImplementedError(
                'async execution of paginated queries is python 3.5.2+ only')

    def __repr__(self):
        return 'paginated({})'.format(self._query)


class Paginator(t.Iterator[T]):
    """An iterator which keeps executing the next query in the page sequece,
    returning the page content."""
    __slots__ = '_executor', '_next_query'

    def __init__(self, next_query, executor):
        self._next_query, self._executor = next_query, executor

    def __iter__(self):
        return self

    def __next__(self):
        if self._next_query is None:
            raise StopIteration()
        page = self._executor(self._next_query)
        self._next_query = page.next_query
        return page.content

    if not PY3:  # pragma: no cover
        next = __next__



"""python 2/3 compatible imports"""
import sys
from operator import attrgetter

PY3 = sys.version_info > (3, )


if PY3:
    import urllib.request as urllib_request
    event_loop = __import__('asyncio').get_event_loop()
    from functools import singledispatch
    from urllib.parse import urlencode
    from urllib.error import HTTPError as urllib_http_error_cls
    from collections.abc import Mapping

    def set_urllib_method(req, method):
        req.method = method

    def func_from_method(method):
        return method

else:  # pragma: no cover
    import urllib2 as urllib_request  # noqa
    from singledispatch import singledispatch  # noqa
    event_loop = None  # noqa
    from urllib import urlencode  # noqa
    func_from_method = attrgetter('im_func')
    urllib_http_error_cls = urllib_request.HTTPError  # noqa
    from collections import Mapping  # noqa

    def set_urllib_method(req, method):
        req.get_method = lambda: method



"""Patch in async-exclusive functionality. Only for python 3+"""
import asyncio
import sys
import typing as t
import urllib.request
from functools import partial
from http.client import HTTPResponse
from io import BytesIO
from itertools import starmap

from .clients import send_async
from .http import Response
from .query import Query

T = t.TypeVar('T')
_ASYNCIO_USER_AGENT = 'Python-asyncio/3.{}'.format(sys.version_info.minor)


class _SocketAdaptor:
    def __init__(self, io):
        self._file = io

    def makefile(self, *args, **kwargs):
        return self._file


@send_async.register(asyncio.AbstractEventLoop)
async def _asyncio_send(loop, req, *, timeout=10, max_redirects=10):
    """A rudimentary HTTP client using :mod:`asyncio`"""
    if not any(h.lower() == 'user-agent' for h in req.headers):
        req = req.with_headers({'User-Agent': _ASYNCIO_USER_AGENT})
    url = urllib.parse.urlsplit(
        req.url + '?' + urllib.parse.urlencode(req.params))
    open_ = partial(asyncio.open_connection, url.hostname, loop=loop)
    connect = open_(443, ssl=True) if url.scheme == 'https' else open_(80)
    reader, writer = await connect
    try:
        headers = '\r\n'.join([
            '{} {} HTTP/1.1'.format(req.method, url.path + '?' + url.query),
            'Host: ' + url.hostname,
            'Connection: close',
            'Content-Length: {}'.format(len(req.content or b'')),
            '\r\n'.join(starmap('{}: {}'.format, req.headers.items())),
        ])
        writer.write(b'\r\n'.join([headers.encode('latin-1'),
                                   b'', req.content or b'']))
        response_bytes = BytesIO(
            await asyncio.wait_for(reader.read(), timeout=timeout))
    finally:
        writer.close()
    resp = HTTPResponse(_SocketAdaptor(response_bytes),
                        method=req.method, url=req.url)
    resp.begin()
    status = resp.getcode()
    if 300 <= status < 400 and 'Location' in resp.headers and max_redirects:
        new_url = urllib.parse.urljoin(req.url, resp.headers['Location'])
        return await _asyncio_send(
            loop, req.replace(url=new_url),
            timeout=timeout,
            max_redirects=max_redirects-1)
    return Response(status, content=resp.read(), headers=resp.headers)


try:
    import aiohttp
except ImportError:  # pragma: no cover
    pass
else:
    @send_async.register(aiohttp.ClientSession)
    async def _aiohttp_send(session, req):
        """send a request with the `aiohttp` library"""
        async with session.request(
                req.method, req.url,
                params=req.params,
                data=req.content,
                headers=req.headers) as resp:
            return Response(resp.status, content=await resp.read(),
                            headers=resp.headers)


@partial(setattr, Query, '__execute_async__')
async def __execute_async__(self, client, auth):
    """Default asynchronous execution logic for a query,
    which uses the query's :meth:`~Query.__iter__`.
    May be overriden for full control of query execution,
    at the cost of reusability.

    .. versionadded:: 1.1

    Note
    ----
    You shouldn't need to override this method, except in rare cases
    where implementing :meth:`Query.__iter__` does not suffice.

    Parameters
    ----------
    client
        the client instance passed to :func:`execute_async`
    auth: ~typing.Callable[[Request], Request]
        a callable to authenticate a :class:`~snug.http.Request`

    Returns
    -------
    T
        the query result
    """
    gen = iter(self)
    request = next(gen)
    while True:
        response = await send_async(client, auth(request))
        try:
            request = gen.send(response)
        except StopIteration as e:
            return e.value


class AsyncPaginator(t.AsyncIterator[T]):
    """An async iterator which keeps executing
    the next query in the page sequence"""
    __slots__ = '_executor', '_next_query'

    def __init__(self, next_query, executor):
        self._next_query, self._executor = next_query, executor

    def __aiter__(self):
        return self

    async def __anext__(self):
        """the content of the next page"""
        if self._next_query is None:
            raise StopAsyncIteration()
        page = await self._executor(self._next_query)
        self._next_query = page.next_query
        return page.content



"""Funtions for dealing with for HTTP clients in a unified manner"""
from .compat import (set_urllib_method, singledispatch, urlencode,
                     urllib_http_error_cls, urllib_request)
from .http import Response

__all__ = ['send', 'send_async']


@singledispatch
def send(client, request):
    """Given a client, send a :class:`~snug.http.Request`,
    returning a :class:`~snug.http.Response`.

    A :func:`~functools.singledispatch` function.

    Parameters
    ----------
    client: any registered client type
        The client with which to send the request.

        Client types registered by default:

        * :class:`urllib.request.OpenerDirector`
          (e.g. from :func:`~urllib.request.build_opener`)
        * :class:`requests.Session`
          (if `requests <http://docs.python-requests.org/>`_ is installed)

    request: Request
        The request to send

    Returns
    -------
    Response
        the resulting response


    Example of registering a new HTTP client:

    >>> @send.register(MyClientClass)
    ... def _send(client, request: Request) -> Response:
    ...     r = client.send(request)
    ...     return Response(r.status, r.read(), headers=r.get_headers())
    """
    raise TypeError('client {!r} not registered'.format(client))


@singledispatch
def send_async(client, request):
    """Given a client, send a :class:`~snug.http.Request`,
    returning an awaitable :class:`~snug.http.Response`.

    A :func:`~functools.singledispatch` function.

    Parameters
    ----------
    client: any registered client type
        The client with which to send the request.

        Client types supported by default:

        * :class:`asyncio.AbstractEventLoop`
          (e.g. from :func:`~asyncio.get_event_loop`)
        * :class:`aiohttp.ClientSession`
          (if `aiohttp <http://aiohttp.readthedocs.io/>`_ is installed)

        Note
        ----
        ``aiohttp`` is only supported on python 3.5.3+

    request: Request
        The request to send

    Returns
    -------
    Response
        the resulting response


    Example of registering a new HTTP client:

    >>> @send_async.register(MyClientClass)
    ... async def _send(client, request: Request) -> Response:
    ...     r = await client.send(request)
    ...     return Response(r.status, r.read(), headers=r.get_headers())
    """
    raise TypeError('client {!r} not registered'.format(client))


@send.register(urllib_request.OpenerDirector)
def _urllib_send(opener, req, **kwargs):
    """Send a request with an :mod:`urllib` opener"""
    if req.content and not any(h.lower() == 'content-type'
                               for h in req.headers):
        req = req.with_headers({'Content-Type': 'application/octet-stream'})
    url = req.url + '?' + urlencode(req.params)
    raw_req = urllib_request.Request(url, req.content, headers=req.headers)
    set_urllib_method(raw_req, req.method)
    try:
        res = opener.open(raw_req, **kwargs)
    except urllib_http_error_cls as http_err:
        res = http_err
    return Response(res.getcode(), content=res.read(), headers=res.headers)


try:
    import requests
except ImportError:  # pragma: no cover
    pass
else:
    @send.register(requests.Session)
    def _requests_send(session, req):
        """send a request with the `requests` library"""
        res = session.request(req.method, req.url,
                              data=req.content,
                              params=req.params,
                              headers=req.headers)
        return Response(res.status_code, res.content, headers=res.headers)



"""
The entire public API is available at root level::

    from snug import Query, Request, send_async, PATCH, paginated, ...
"""
import sys

from . import clients, http  # noqa
from .__about__ import (__author__, __copyright__, __description__,  # noqa
                        __version__)
from .clients import *  # noqa
from .http import *  # noqa
from .pagination import *  # noqa
from .query import *  # noqa

if sys.version_info > (3, ):  # pragma: no cover
    from . import _async  # noqa



import os.path

from setuptools import setup, find_packages


def read(fname):
    path = os.path.join(os.path.dirname(__file__), fname)
    with open(path, 'r') as rfile:
        return rfile.read()


metadata = {}
exec(read('snug/__about__.py'), metadata)


setup(
    name='snug',
    version=metadata['__version__'],
    description=metadata['__description__'],
    license='MIT',
    long_description=read('README.rst') + '\n\n' + read('HISTORY.rst'),
    url='https://github.com/ariebovenberg/snug',

    author=metadata['__author__'],
    author_email='a.c.bovenberg@gmail.com',

    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',

        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    install_requires=[
        'typing>=3.6.6; python_version<"3.5"',
        'singledispatch>=3.4.0; python_version<"3.4"',
    ],
    extras_require={
        'aiohttp': ['aiohttp>=3.4.4,<3.6.0'],
        'requests': ['requests>=2.20,<2.22'],
    },
    keywords=['api-wrapper', 'http', 'generators', 'async',
              'graphql', 'rest', 'rpc'],
    python_requires='>=2.7',
    packages=find_packages(exclude=('examples', 'tests', 'docs', 'tutorial')),
)



"""Basic HTTP abstractions and functionality"""
from base64 import b64encode
from functools import partial
from itertools import chain
from operator import attrgetter, methodcaller

from .compat import Mapping

__all__ = [
    'Request',
    'Response',
    'header_adder',
    'prefix_adder',
    'basic_auth',
    'GET',
    'POST',
    'PUT',
    'PATCH',
    'DELETE',
    'HEAD',
    'OPTIONS',
]


class _FrozenDict(Mapping):
    __slots__ = '_inner'

    def __init__(self, inner=()):
        self._inner = dict(inner)

    __len__ = property(attrgetter('_inner.__len__'))
    __iter__ = property(attrgetter('_inner.__iter__'))
    __getitem__ = property(attrgetter('_inner.__getitem__'))
    __repr__ = property(attrgetter('_inner.__repr__'))


class _SlotsMixin(object):
    __slots__ = ()

    def _asdict(self):
        return {a: getattr(self, a) for a in self.__slots__}

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._asdict() == other._asdict()
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, self.__class__):
            return self._asdict() != other._asdict()
        return NotImplemented

    def replace(self, **kwargs):
        """Create a copy with replaced fields

        Parameters
        ----------
        **kwargs
            fields and values to replace
        """
        return type(self)(**_merge_maps(self._asdict(), kwargs))


def _merge_maps(m1, m2):
    """merge two Mapping objects, keeping the type of the first mapping"""
    return type(m1)(chain(m1.items(), m2.items()))


class Request(_SlotsMixin):
    """A simple HTTP request.

    Parameters
    ----------
    method: str
        The http method
    url: str
        The requested url
    content: bytes or None
        The request content
    params: Mapping
        The query parameters.
    headers: Mapping
        Request headers.
    """
    __slots__ = 'method', 'url', 'content', 'params', 'headers'
    __hash__ = None

    def __init__(self, method, url, content=None,
                 params=_FrozenDict(), headers=_FrozenDict()):
        self.method = method
        self.url = url
        self.content = content
        self.params = params
        self.headers = headers

    def with_headers(self, headers):
        """Create a new request with added headers

        Parameters
        ----------
        headers: Mapping
            the headers to add
        """
        return self.replace(headers=_merge_maps(self.headers, headers))

    def with_prefix(self, prefix):
        """Create a new request with added url prefix

        Parameters
        ----------
        prefix: str
            the URL prefix
        """
        return self.replace(url=prefix + self.url)

    def with_params(self, params):
        """Create a new request with added query parameters

        Parameters
        ----------
        params: Mapping
            the query parameters to add
        """
        return self.replace(params=_merge_maps(self.params, params))

    def __repr__(self):
        return ('<Request: {0.method} {0.url}, params={0.params!r}, '
                'headers={0.headers!r}>').format(self)


class Response(_SlotsMixin):
    """A simple HTTP response.

    Parameters
    ----------
    status_code: int
        The HTTP status code
    content: bytes or None
        The response content
    headers: Mapping
        The headers of the response.
    """
    __slots__ = 'status_code', 'content', 'headers'
    __hash__ = None

    def __init__(self, status_code, content=None, headers=_FrozenDict()):
        self.status_code = status_code
        self.content = content
        self.headers = headers

    def __repr__(self):
        return ('<Response: {0.status_code}, '
                'headers={0.headers!r}>').format(self)


def basic_auth(credentials):
    """Create an HTTP basic authentication callable

    Parameters
    ----------
    credentials: ~typing.Tuple[str, str]
        The (username, password)-tuple

    Returns
    -------
    ~typing.Callable[[Request], Request]
        A callable which adds basic authentication to a :class:`Request`.
    """
    encoded = b64encode(':'.join(credentials).encode('ascii')).decode()
    return header_adder({'Authorization': 'Basic ' + encoded})


prefix_adder = partial(methodcaller, 'with_prefix')
prefix_adder.__doc__ = """
Make a callable which adds a prefix to a request url

Example
-------

>>> func = snug.prefix_adder('https://api.test.com/v1/')
>>> func(snug.GET('foo/bar/')).url
https://api.test.com/v1/foo/bar/
"""
header_adder = partial(methodcaller, 'with_headers')
header_adder.__doc__ = """
Make a callable which adds headers to a request

Example
-------

>>> func = snug.header_adder({'content-type': 'application/json'})
>>> func(snug.GET('https://test.dev')).headers
{'content-type': 'application/json'}
"""
GET = partial(Request, 'GET')
GET.__doc__ = "Shortcut for a GET request"
POST = partial(Request, 'POST')
POST.__doc__ = "Shortcut for a POST request"
PUT = partial(Request, 'PUT')
PUT.__doc__ = "Shortcut for a PUT request"
PATCH = partial(Request, 'PATCH')
PATCH.__doc__ = "Shortcut for a PATCH request"
DELETE = partial(Request, 'DELETE')
DELETE.__doc__ = "shortcut for a DELETE request"
HEAD = partial(Request, 'HEAD')
HEAD.__doc__ = "Shortcut for a HEAD request"
OPTIONS = partial(Request, 'OPTIONS')
OPTIONS.__doc__ = "Shortcut for a OPTIONS request"
